-- Puffin: A Tojaso FowlPlay Project
-- Author: Tomber, contact tomber@toja.me
-- Copyright 2016, All rights reserved

-- This file contains Puffin's database manager which is used to share information between components.
-- In practice, model components react to game events and update a representation of the game's state in the database,
-- adapter components process that information (e.g., by sorting and filtering it), and view
-- components display the processed information. The database is also used to store configuration info.

-- Data management functions are based on a three-level hierarchy: database, record, attribute.
-- Attributes are name-value pairs (values can be tables with memory managed by caller).
-- Records can be redirected to access attributes from a caller-provided table.
-- Optionally, a validate function can be called before accessing a data field to make sure it is valid.
-- Databases can be tagged as persistent so that they will be saved and restored between sessions.

local _, Puffin = ...
local allocate = Puffin.AllocateTable
local release = Puffin.ReleaseTable
local copyTable = Puffin.CopyTable

local Puffin_Data = {}
local databases = {}
local empty = {}

-- Allocate a new database (optionally wipe if already exists), initialize records and set persistent tag
function Puffin_Data.AllocateDatabase(name, persistent, wipe)
	local dt = databases[name]
	if dt then
		if wipe then
			for k, v in pairs(dt._records) do dt._records[k] = release(v) end
			table.wipe(dt._redirect)
		end
	else
		dt = allocate(); dt._records = allocate(); dt._redirect = allocate(); databases[name] = dt
	end
	dt._persistent = persistent
end

-- Return true if the named database exists
function Puffin_Data.IsDatabase(name)
	local dt = databases[name]
	return dt ~= nil
end

-- Release and recycle a database, release all tables used to store records and attributes
function Puffin_Data.ReleaseDatabase(name)
	local dt = databases[name]
	if dt then
		for k, v in pairs(dt._records) do dt._records[k] = release(v) end
		release(dt._records); release(dt._redirect); release(dt)
	end
	databases[name] = nil
end

-- Release all current records in a database
function Puffin_Data.ReleaseRecords(name)
	local dt = databases[name]
	if dt then
		for k, v in pairs(dt._records) do dt._records[k] = release(v) end
		release(dt._redirect)
	end
end

-- Return values equivalent to pairs() that can be used to iterate over all non-redirected records in the database
function Puffin_Data.GetRecords(name)
	local dt = databases[name]
	if dt then return pairs(dt._records) end
	return pairs(empty)
end

-- Set attributes in a database record, optionally wiping current contents before updating
function Puffin_Data.SetRecord(name, record, atable, wipe)
	local dt = databases[name]
	if dt and record and atable then
		local dr = dt._records[record]
		if not dr then dr = allocate(); dt._records[record] = dr end
		if wipe then table.wipe(dr) end -- first delete current contents
		for key, value in pairs(atable) do dr[key] = value end
		dt._redirect[record] = nil -- removing any redirect set for this record
	end
end

-- Set a redirection table for a record, attributes will be read from this table instead of the record's internal table
-- Memory management for the redirect table is the responsibility of the caller
function Puffin_Data.SetRedirect(name, record, atable)
	local dt = databases[name]
	if dt and record and atable then
		local dr = dt._records[record]
		if dr then dt._records[record] = release(dr) end --release old record attribute table
		dt._redirect[record] = atable
	end
end

-- Set a validation function for a database to be called before returning a record or attribute
function Puffin_Data.SetValidate(name, validate)
	local dt = databases[name]
	if dt and type(validate) == "function" then dt._validate = validate end
end

-- Get the attributes in a database record
function Puffin_Data.GetRecord(name, record)
	local dt = databases[name]
	if dt and record then
		if dt._validate then dt._validate(name, record) end -- optional validate function called before returning the record
		local dr = dt._records[record]; if not dr then dr = dt._redirect[record] end
		return dr
	end
	return nil
end

-- Release a database record, return true only if there was a record to be released
function Puffin_Data.ReleaseRecord(name, record)
	local dt = databases[name]
	if dt and record then
		local dr = dt._records[record]; if dr then dt._records[record] = release(dr); return true end
		dr = dt._redirect[record]; if dr then dt._redirect[record] = nil; return true end
	end
	return nil
end

-- Compare all entries in a table to attributes in a record (ignore extra attributes in the record)
function Puffin_Data.CompareRecord(name, record, atable)
	local dt = databases[name]
	if dt and record then
		local dr = dt._records[record]; if not dr then dr = dt._redirect[record] end
		if dr then for key, value in pairs(atable) do if dr[key] ~= value then return false end end return true end
	end
	return false
end

-- Set an attribute's value in a database record, following redirect if any and creating a record table if necessary
function Puffin_Data.SetAttribute(name, record, attribute, value)
	local dt = databases[name]
	if dt and record then
		local dr = dt._records[record]
		if not dr then dr = dt._redirect[record] end -- use redirect table if set
		if not dr then dr = allocate(); dt._records[record] = dr end
		dr[attribute] = value
	end
end

-- Set multiple attributes in the same record
function Puffin_Data.SetAttributes(name, record, atable)
	local dt = databases[name]
	if dt and record then
		local dr = dt._records[record]
		if not dr then dr = dt._redirect[record] end -- use redirect table if set
		if not dr then dr = allocate(); dt._records[record] = dr end
		for attribute, value in pairs(atable) do dr[attribute] = value end
	end
end

-- Get the value of an attribute in a database record, following redirect if any
function Puffin_Data.GetAttribute(name, record, attribute)
	local dt = databases[name]
	if dt and record then
		if dt._validate then dt._validate(name, record, attribute) end -- optional validate function called before returning the attribute
		local dr = dt._records[record]; if not dr then dr = dt._redirect[record] end -- use redirect table if set
		if dr then return dr[attribute] end
	end
	return nil
end

-- Save persistent databases to a profile just before shutdown (okay to use non-memory managed copy)
function Puffin_Data.SavePersistentData()
	local profile = {}
	for name, dt in pairs(databases) do -- walk thru all databases and find databases with a persistent tag
		if dt._persistent then
			local ct = {} -- need to make a local copy of database records
			for record, dr in pairs(dt._records) do ct[record] = copyTable(dr) end -- copy the records
			for record, dr in pairs(dt._redirect) do ct[record] = copyTable(dr) end -- copy redirected records (experimental)
			profile[name] = ct
		end
	end
	PuffinDB = profile -- SharedVariables table automatically saved by the client
end

-- Restore persistent databases from a profile, assume this is done during initialization with database starting empty
function Puffin_Data.RestorePersistentData()
	local profile = PuffinDB -- SharedVariables table automatically restored by the client
	if profile and type(profile) == "table" then
		for name, atable in pairs(profile) do -- copy databases from the profile and create databases with the persistent tag
			local dt = allocate()
			dt._records = allocate()
			dt._redirect = allocate()
			dt._persistent = true
			databases[name] = dt
			if atable and type(atable) == "table" then
				for record, attributes in pairs(atable) do
					if attributes and type(attributes) == "table" then
						local dr = allocate()
						for key, value in pairs(attributes) do dr[key] = copyTable(value) end -- deep copy values from the profile
						dt._records[record] = dr
					end
				end
			end
		end
	end
end

Puffin.Database = Puffin_Data -- this reference is used for the save/restore function
