-- Puffin: A Tojaso FowlPlay Project
-- Author: Tomber, contact tomber@toja.me
-- Copyright 2016, All rights reserved

-- This file implements a framework that supports model, view and adapter objects which interact using a shared database and events.
-- The framework provides memory management, callbacks for events, timers, update and refresh cycles, and error handling.

local _, Puffin = ...
Puffin_Addon = Puffin -- only one variable added to the global name space (required for access from load-on-demand options)
Puffin.enabled = true -- enable/disable the addon

-- Error handling support for reporting problems
function Puffin.Error(source, errmsg) return "Puffin (" .. source .. "): " .. errmsg end

-- Display debug messages with variable number of arguments in a useful format
function Puffin.Debug(a, ...)
	if type(a) == "table" then
		for k, v in pairs(a) do print(tostring(k) .. " = " .. tostring(v)) end -- if first parameter is a table, print out its fields
	else
		local s = tostring(a) -- otherwise first argument is a string but just make sure
		local parm = {...}
		for i = 1, #parm do s = s .. " " .. tostring(parm[i]) end -- append remaining arguments converted to strings
		print(s)
	end
end

-- Memory management
local tablePool = {} -- pool of available tables
function Puffin.AllocateTable() local b = next(tablePool); if b then tablePool[b] = nil else b = {} end return b end
function Puffin.ReleaseTable(b) if b then table.wipe(b); tablePool[b] = true end return nil end

-- Copy a table, including its metatable
function Puffin.CopyTable(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

-- Event callback handling
local registered = {} -- table of all registered callbacks
local callbackNumber = 0 -- unique id for each registered callback
local callbacks = {} -- table of tables for event callbacks
local called = false -- set to true when any callback is triggered
local eventFrame = nil -- frame that is only used for event handling

-- Register a callback for an event (multiple callbacks can be registered for the same event), return a unique id
-- The unique id can be used to suspend, resume, and unregister the callback
-- Callbacks are functions with a single optional argument (if not specified then event is passed)
-- Note that any time an event results in a callback the value of "called" is set to true
function Puffin.RegisterCallback(event, f, arg, suspend)
	if not event or type(f) ~= "function" then return nil end
	local isSoft = string.find(event, "Puffin") -- software events all include the string "Puffin"
	if not isSoft and not eventFrame then
		eventFrame = CreateFrame("Frame")
		function eventFrame:OnEvent(event, ...)
			local cbs = callbacks[event]
			if cbs then for _, cb in pairs(cbs) do (cb.f)(cb.arg, ...); called = true end end
		end
		eventFrame:SetScript("OnEvent", eventFrame.OnEvent)
	end
	callbackNumber = callbackNumber + 1
	local cb = Puffin.AllocateTable()
	cb.event = event; cb.f = f; cb.arg = arg or event
	registered[callbackNumber] = cb -- table contains all registered callbacks, indexed by unique id
	local cbs = callbacks[event]
	if not cbs then
		cbs = Puffin.AllocateTable()
		callbacks[event] = cbs -- new event so have to allocate a table to hold callbacks
		if not isSoft then eventFrame:RegisterEvent(event) end -- register the frame to receive a hardware event
	end
	if not suspend then cbs[callbackNumber] = cb end -- add to event-specific table of callbacks
	return callbackNumber
end

-- Suspend sending events to a specified callback id
function Puffin.SuspendCallback(id)
	local cb = registered[id]
	if cb then
		local cbs = callbacks[cb.event]
		if cbs then cbs[id] = nil end -- remove from event-specific table of callbacks
	end
end

-- Resume sending events to a specified callback id
function Puffin.ResumeCallback(id)
	local cb = registered[id]
	if cb then
		local event = cb.event
		local cbs = callbacks[event]
		if not cbs then
			cbs = Puffin.AllocateTable()
			callbacks[event] = cbs -- ressumed event so have to allocate another table to hold callbacks
			local isSoft = string.find(event, "Puffin") -- software events all include the string "Puffin"
			if not isSoft then eventFrame:RegisterEvent(event) end -- register the frame to receive a hardware event
		end
		cbs[id] = cb -- add to event-specific table of callbacks
	end
end

-- Unregister a callback for a specified callback id
function Puffin.UnregisterCallback(id)
	local cb = registered[id]
	if cb then
		local event = cb.event
		local cbs = callbacks[event] -- update event-specific table
		if cbs then
			cbs[id] = nil -- remove from active callbacks
			if not next(cbs) then -- check if the event-specific table is now empty
				callbacks[event] = Puffin.ReleaseTable(cbs) -- if so then release the table...
				local isSoft = string.find(event, "Puffin") -- software events all include the string "Puffin"
				if not isSoft then eventFrame:UnregisterEvent(event) end -- and, if not software event, unregister from frame
			end
		end
		registered[id] = Puffin.ReleaseTable(cb) -- remove from table of all callbacks as well
	end
end

-- Trigger any callbacks for the specified event, if any, passing along optional parameters
function Puffin.Callback(event, ...) eventFrame:OnEvent(event, ...) end

-- Return true if any callbacks are registered for the specified event
function Puffin.CallbackRegistered(event) return event and callbacks[event] and next(callbacks[event]) end

-- Timers
local timers = {}
local timed = false --set to true when any timer is triggered
function Puffin.StartTimer(event, timeLeft) timers[event] = GetTime() + timeLeft end

-- Update management
Puffin.updateTarget = 0.1 -- default update rate is 10 times per second
local lastTime, updateTime, refreshTime = 0, 0, 0 -- time since last update and refresh
local forced = false -- set to force an immediate update, obviously you don't want to do this very often
function Puffin.ForceUpdate() forced = true end

-- Handle update calls generated by the game, throttle to minimize performance impact, turn into either Update or Refresh callbacks
-- depending on whether any events have been triggered
local function UpdateHandler()
	local now = GetTime(); local elapsed = now - lastTime; lastTime = now
	for event, t in pairs(timers) do if t < now then timers[event] = nil; Puffin.Callback(event); timed = true end end
	
	if elapsed > 1.0 then elapsed = 1.0 end -- should only happen during initialization
	
	if updateTime < 0 then updateTime = elapsed else updateTime = updateTime + elapsed end -- timer for update cycles
	if forced or (updateTime >= Puffin.updateTarget) then -- limit update rate unless forced by software override
		if called or timed or forced then -- only do major updates when events warrant it
			called = false; timed = false; forced = false
			Puffin.Callback("Puffin_Update_Priority"); Puffin.Callback("Puffin_Update")
		end
		updateTime = updateTime - Puffin.updateTarget
	end
	
	Puffin.Callback("Puffin_Refresh_Priority"); Puffin.Callback("Puffin_Refresh")
	C_Timer.After(0.001, UpdateHandler) -- register to be called for next frame
end

function Puffin.ADDON_LOADED(e, name) if name == "Puffin" then local db = Puffin.Database; if db then db.RestorePersistentData() end Puffin.InitializeLDB() end end
function Puffin.PLAYER_LOGIN() Puffin.Callback("Puffin_Initialize_Priority"); Puffin.Callback("Puffin_Initialize") end
function Puffin.PLAYER_ENTERING_WORLD() Puffin.Callback("Puffin_Update_Priority"); Puffin.Callback("Puffin_Update"); C_Timer.After(0.001, UpdateHandler) end
function Puffin.PLAYER_LOGOUT() local db = Puffin.Database; if db then db.SavePersistentData() end end

Puffin.RegisterCallback("ADDON_LOADED", Puffin.ADDON_LOADED) -- fires when addon loaded and saved variables have been read from file
Puffin.RegisterCallback("PLAYER_LOGIN", Puffin.PLAYER_LOGIN) -- signals most game info is available and addon initialization can happen
Puffin.RegisterCallback("PLAYER_ENTERING_WORLD", Puffin.PLAYER_ENTERING_WORLD) -- fires whenever player sees a loading screen
Puffin.RegisterCallback("PLAYER_LOGOUT", Puffin.PLAYER_LOGOUT) -- fires when player exits and just before saved variables are written to file

-- Use LibDataBroker to create a launcher icon
Puffin.minimapSettings = { hide = false, minimapPos = 180, radius = 80, }
function Puffin.InitializeLDB()
	local libLDB = LibStub("LibDataBroker-1.1", true)
	if libLDB then
		local ldb = libLDB:NewDataObject("Puffin", {
			type = "launcher", text = "Puffin", icon = [[Interface\AddOns\Puffin\Media\Puffin]],
			OnClick = function(_, msg) if msg == "LeftButton" then Puffin.OptionsPanel() end end,
			OnTooltipShow = function(tooltip) if not tooltip or not tooltip.AddLine then return end tooltip:AddLine("Puffin") end,
		})
		if ldb then
			Puffin.ldbi = LibStub("LibDBIcon-1.0", true)
			if Puffin.ldbi then Puffin.ldbi:Register("Puffin", ldb, minimapSettings) end
		end
	end
end

-- Launch the load-on-demand options panel
local optionsLoaded, optionsFailed
function Puffin.OptionsPanel()
	if not optionsLoaded then
		optionsLoaded = true
		local loaded, reason = LoadAddOn("Puffin_Options")
		if not loaded then print("Failed to load Puffin Options: " .. tostring(reason)); optionsFailed = true end
	end
	if not optionsFailed then Puffin.ToggleOptions() end
end
