-- Puffin: A Tojaso FowlPlay Project
-- Widget base class defines common characteristics for all display widgets
-- Implements pixel perfect alignment based on display resolution and scale

-- Shared attributes:
-- width, height - dimensions of the widget's container that is parent frame for display elements
-- scale - factor by which to increase/decrease the size of the widget
-- strata - frame strata for the widget
-- level - base frame level for the widget
-- depth - relative frame level for the widget (higher shows on top of lower)
-- alpha - factor by which to increase/decrease opacity of the widget
-- hidden - boolean to toggle hiding the widget
-- condition - boolean that results in hiding the widget when exactly equal to false

local _, Puffin = ...
local EF = Puffin.Error
local media = LibStub("LibSharedMedia-3.0")
media:Register("font", "Roboto Medium", [[Interface\Addons\Puffin\media\Roboto-Medium.ttf]])
media:Register("font", "Roboto Regular", [[Interface\Addons\Puffin\media\Roboto-Regular.ttf]])
media:Register("font", "Roboto Bold", [[Interface\Addons\Puffin\media\Roboto-Bold.ttf]])
media:Register("font", "Roboto Italic", [[Interface\Addons\Puffin\media\Roboto-Italic.ttf]])

local class = Puffin.Class
local defaultWhite = { r = 1, g = 1, b = 1, a = 1 }
local defaultBlack = { r = 0, g = 0, b = 0, a = 1 }
local defaultBlue = { r = 0, g = 0, b = 1, a = 1 }

local Widget = class("Widget")
Puffin.Classes.Widget = Widget

local pixelScale, pixelWidth, pixelHeight, worldScale -- screen properties used for pixel perfect alignment

-- Validate a resolution string with format "w x h" as used by the system menu for selecting display settings
local function ValidResolution(res)
	if type(res) == "string" then
		local w, h = DecodeResolution(res)
		if w and h and type(w) == "number" and type(h) == "number" then
			if w > 0 and h > 0 then return true end
		end
	end
	return false
end

-- Initialize values needed for pixel perfect alignment
local function InitializeAlign()
	local monitorIndex = Display_PrimaryMonitorDropDown:GetValue() -- get current monitor index (should be same as the cvar "gxMonitor")
	local isWindowed = Display_DisplayModeDropDown:windowedmode() -- test if in windowed mode (used to fallback to cvar value for resolution)
	local isFullscreen = Display_DisplayModeDropDown:fullscreenmode() -- test if in fullscreen mode (used to fallback to cvar value for resolution)
	local resolutionIndex = GetCurrentResolution(monitorIndex) -- get index for current resolution in list of screen resolutions
	if not resolutionIndex then resolutionIndex = 0 end -- make sure valid number for next test...
	local resolution = resolutionIndex > 0 and select(resolutionIndex, GetScreenResolutions(monitorIndex)) or nil -- best case scenario for accurate resolution
	if not ValidResolution(resolution) then resolution = isFullscreen and GetCVar("gxFullscreenResolution") or GetCVar("gxWindowedResolution") end
	if ValidResolution(resolution) then -- should have valid resolution at this point, either from screen resolutions or appropriate cvar
		pixelWidth, pixelHeight = DecodeResolution(resolution) -- use Blizzard's utility function to decode the resolution width and height
		pixelScale = GetScreenHeight() / pixelHeight -- figure out how big virtual pixels are versus screen pixels
	else
		pixelWidth = GetScreenWidth(); pixelHeight = GetScreenHeight() -- ultimate fallback safe values for width and height
		pixelScale = 1 -- and safe value for pixel perfect calculations
	end
	worldScale =  GetScreenHeight() / 768 -- used when converting from UIParent to WorldFrame coordinates
	-- Puffin.Debug("resolution", monitorIndex, resolutionIndex, resolution, ValidResolution(resolution), pixelWidth, pixelHeight, pixelScale)
end

Puffin.RegisterCallback("Puffin_Initialize_Priority", InitializeAlign) -- routine to calculate pixel perfect alignment factor

local function Pixel(x) return x * pixelScale end -- convert from screen units to pixels
local function Screen(x) return x / pixelScale end -- convert from pixel units to screen
local function World(x) return x * worldScale end -- convert from screen to world units
local function Align(x) return pixelScale * math.floor((x / pixelScale) + 0.5) end
Puffin.Pixel = Pixel; Puffin.Screen = Screen; Puffin.World = World; Puffin.Align = Align

-- Manage the callback suspend/resume status for an object
function Puffin.ManageCallback(w, active)
	if active then
		if not w._callback_active then Puffin.ResumeCallback(w._callback); w._callback_active = true end
	else
		if w._callback_active then Puffin.SuspendCallback(w._callback); w._callback_active = false end
	end
end

-- Widget class constructor sets default position and size attributes, registers display frames from subclasses
-- Optional template includes default values for widget class, optional layout includes values for new instance
function Widget:initialize(frames, template, layout)
	local tc = frames.container; assert(tc, EF("Widget_Class", "undefined container"))
	if not self.scale then self.scale = 1 end
	if not self.depth then self.depth = 1 end
	if not self.alpha then self.alpha = 1 end
	self.hidden = false
	self.condition = true
	self.width = 0; self.height = 0
	self._display = frames -- display frames are created by subclasses
	tc:EnableMouse(false) -- default is not to interact with mouse
	self:Configure(template, layout) -- apply template and layout
end

-- Apply template and layout values to a widget
function Widget:Configure(template, layout)
	if template then for k, v in pairs(template) do self[k] = v end end
	if layout then for k, v in pairs(layout) do self[k] = v end end
end

-- Data access for this class with both generic and convenience methods
function Widget:Set(k, v) self[k] = v end -- generic method, override in subclasses to track which values get changed
function Widget:SetSize(w, h) self.width = w; self.height = h end
function Widget:SetScale(s) self.scale = s end
function Widget:SetAlpha(a) self.alpha = a end
function Widget:SetStrata(s) self.strata = s end
function Widget:SetLevel(l) self.level = l end
function Widget:SetDepth(d) self.depth = d end
function Widget:SetHidden(h) self.hidden = h end
function Widget:SetCondition(c) self.condition = c end
function Widget:Get(k) return self[k] end -- generic method, override in subclasses if need to compute values

-- Set display location of the widget, relative to a specified frame or another widget or to the screen
function Widget:SetPosition(point, widget, relative, offsetX, offsetY)
	local t = self._display; assert(t, EF("Widget_Class", "undefined widget"))
	local tc = t.container; assert(tc, EF("Widget_Class", "undefined container"))
	local frame = widget or UIParent
	if widget and widget._display and widget._display.container then frame = widget._display.container end
	tc:ClearAllPoints(); tc:SetPoint(point, frame, relative, Pixel(offsetX or 0), Pixel(offsetY or 0))
	if frame == UIParent then -- reposition any widget anchored to UIParent for pixel perfect alignment
		if self.width and self.height then tc:SetSize(Pixel(self.width), Pixel(self.height)) end -- must have dimensions set
		local left, bottom = tc:GetLeft(), tc:GetBottom()
		if left and bottom then tc:ClearAllPoints(); tc:SetPoint("BOTTOMLEFT", nil, "BOTTOMLEFT", Align(left), Align(bottom)) end
	end
end

-- Set the widget's parent widget to establish the hierarchy for inherited properties such as scale and alpha
function Widget:SetParent(widget)
	local t = self._display; assert(t, EF("Widget_Class", "undefined widget"))
	local tc = t.container; assert(tc, EF("Widget_Class", "undefined container"))
	local frame = UIParent
	if widget and widget._display and widget._display.container then frame = widget._display.container end
	tc:SetParent(frame)
	if not self.depth then tc:SetFrameLevel(frame:GetFrameLevel() + 1) end
end

-- Update the widget when data attributes have changed
function Widget:Update()
	local t = self._display; assert(t, EF("Widget_Class", "undefined widget"))
	local tc = t.container; assert(tc, EF("Widget_Class", "undefined container"))
	assert(self.width and self.width > 0, EF("Widget_Class", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Class", "undefined height"))
	if self.hidden or (self.condition == false) then
		tc:Hide()
	else
		local w, h = Pixel(self.width), Pixel(self.height)
		if w ~= self._width or h ~= self._height then tc:SetSize(w, h); self._width = w; self._height = h end
		local x = self.alpha or 1; if self._alpha ~= x then tc:SetAlpha(x); self._alpha = x end
		x = self.scale or 1; if self_scale ~= x then tc:SetScale(x); self._scale = x end
		x = self.depth + (self.level or 0); if self._depth ~= x then tc:SetFrameLevel(x); self._depth = x end
		x = self.strata or "MEDIUM"; if self._strata ~= x then tc:SetFrameStrata(x); self._strata = x end
		tc:Show()
	end
end

-- Release resources and hide display elements when widget no longer needed (placeholder for subclasses)
function Widget:Release() end

-- Widget_Backdrop: abstract class that adds a backdrop and border

-- Custom attributes:
-- backdropTexture - if defined then show a background texture
-- backdropColor - color to shade the backdrop texture
-- backdropTile - if defined then size of tile in pixels for repeating background texture
-- backdropInset - distance from edge of frame to backdrop texture (default: 2)
-- borderTexture - if defined then show a border around the widget
-- borderColor - color to apply to border (default: black)
-- borderWidth - width of border (default: 8)
-- borderOffset - distance of border from widget's edge (default: 2)

local Widget_Backdrop = Widget:subclass("Widget_Backdrop")
Puffin.Classes.Widget_Backdrop = Widget_Backdrop

-- Widget class constructor sets default position and size attributes, registers display frames from subclasses
-- Optional template includes default values for widget class, optional layout includes values for new instance
function Widget_Backdrop:initialize(frames, template, layout)
	self.backdropTexture = nil; self.backdropTile = nil; self.backdropColor = nil; self.backdropInset = nil
	self.borderTexture = nil; self.borderColor = nil; self.borderWidth = nil; self.borderOffset = nil
	self.backdrop = { tile = false, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 }}
	Widget.initialize(self, frames, template, layout) -- initialize the superclass
end

-- Data access for this class with both generic and convenience methods
function Widget_Backdrop:SetBackdrop(tex, c, t, o) self.backdropTexture = tex; self.backdropColor = c; self.backdropTile = t; self.backdropInset = o end
function Widget_Backdrop:SetBorder(t, c, w, o) self.borderTexture = t; self.borderColor = c; self.borderWidth = w; self.borderOffset = o end

-- Update the widget when data attributes have changed
function Widget_Backdrop:Update()
	Widget.Update(self) -- first do update in the superclass
	local t = self._display
	local tc = t.container
	if tc:IsShown() then
		if self.backdropTexture or self.borderTexture then -- check if showing a backdrop and/or border
			local b = self.backdrop
			local tex = self.borderTexture
			if tex then mt = media:Fetch("border", tex, true); if mt or tex == "None" then tex = mt end end -- check if border texture is in LibSharedMedia
			b.edgeFile = tex
			tex = self.backdropTexture
			if tex then mt = media:Fetch("background", tex, true); if mt or tex == "None" then tex = mt end end -- check if background texture is in LibSharedMedia
			b.bgFile = tex
			local x = Pixel(self.backdropInset or 2)
			local insets = b.insets; insets.left = x; insets.right = x; insets.top = x; insets.bottom = x
			b.edgeSize = Pixel(self.borderWidth or 8); if (b.edgeSize < Pixel(1)) then b.edgeSize = Pixel(1) end
			if self.backdropTile and self.backdropTile > 0 then b.tile = true; b.tileSize = self.backdropTile else b.tile = false; b.tileSize = 0 end
			local offset = Pixel(2 * (self.borderOffset or 2))
			local tb = t.backdrop
			if not tb then tb = CreateFrame("Frame", nil, tc); t.backdrop = tb end -- create backdrop frame if needed
			tb:SetSize(tc:GetWidth() + offset, tc:GetHeight() + offset)
			local level = tc:GetFrameLevel()
			if tex then level = level - 1 else level = level + 1 end -- if no background texture then put border on top
			tb:SetFrameLevel(level); tb:ClearAllPoints(); tb:SetPoint("CENTER", tc, "CENTER"); tb:SetBackdrop(b)
			local c = self.backdropColor or defaultWhite; tb:SetBackdropColor(c.r, c.g, c.b, c.a or 1)			
			c = self.borderColor or defaultBlack; tb:SetBackdropBorderColor(c.r, c.g, c.b, c.a or 1)
		else
			if t.backdrop then t.backdrop:SetBackdrop(nil) end
		end
	end
end

-- Widget_Panel: display a rectangular panel with a border
-- This basically just allocates the display objects needed by the widget superclass

local Widget_Panel = Widget_Backdrop:subclass("Widget_Panel")
Puffin.Classes.Widget_Panel = Widget_Panel

local recycledPanels = {}
local panelNumber = 0

-- Create a new panel widget
function Widget_Panel:initialize(template, layout)
	local n, panel = next(recycledPanels) -- get any available recycled set of panel frames
	if n then
		recycledPanels[n] = nil
	else
		panel = {}
		panelNumber = panelNumber + 1
		local frameName = "Puffin_Panel" .. tostring(panelNumber) -- generate unique name for frames
		panel.container = CreateFrame("Frame", frameName .. "Panel", UIParent)
	end
	Widget_Backdrop.initialize(self, panel, template, layout) -- pass table of display frames used by this widget plus optional template and layout
end

-- Release panel widget when no longer needed, recycle display objects associated with it
function Widget_Panel:Release()
	local t = self._display; assert(t, EF("Widget_Panel", "undefined panel"))
	t.container:Hide()
	table.insert(recycledPanels, t)
	self._display = nil
end

-- Widget to show text in a variety of formats (including countdown time)

-- Custom attributes:
-- font - font to use for the text (default: ChatFontNormal)
-- fsize - font size for the text (default: 10)
-- delta - delta to apply to font size (default: 0)
-- shadow - enable a shadow behind the text (default: false)
-- outline - enable font "OUTLINE" flag (default: false)
-- thick - enable font "THICKOUTLINE" flag (default: false)
-- mono - enable font "MONOCHROME" flag (default: false)
-- wrap - enable wrap of text to multiple lines (default: false)
-- justifyH - horizontal text alignment (default: "CENTER")
-- justifyV - vertical text alignment (default: "MIDDLE")
-- text - string to display when textFormat is "text" (default: "")
-- color - color to apply to text (default: none)
-- offsetX - horizontal displacement of text (default: 0)
-- offsetY - vertical displacement of text (default: 0)

local Widget_Text = Widget:subclass("Widget_Text")
Puffin.Classes.Widget_Text = Widget_Text

local recycledTexts = {}
local textNumber = 0

-- Create a new text widget object
function Widget_Text:initialize(template, layout)
	local n, text = next(recycledTexts) -- get any available recycled widget
	if n then
		recycledTexts[n] = nil
	else
		text = {}
		textNumber = textNumber + 1
		local frameName = "Puffin_Text" .. tostring(textNumber) -- generate unique name for text's frame
		text.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		text.fontstring = text.container:CreateFontString(nil, "ARTWORK", ChatFontNormal)
	end
	self.changedFont = true; self.changedAlignment = true
	self.font = nil; self.fsize = nil; self.delta = nil; self.shadow = nil; self.outline = nil; self.thick = nil; self.mono = nil -- clear custom attributes
	self.justifyH = nil; self.justifyV = nil; self.wrap = nil; self.color = nil; self.text = nil
	Widget.initialize(self, text, template, layout) -- pass table of display frames used by this widget plus optional template and layout
end

-- Configure with any changes to template and layout
function Widget_Text:Configure(template, layout)
	self.changedFont = true; self.changedAlignment = true
	Widget.Configure(self, template, layout)
end

-- Data access methods for setting text-related attributes
function Widget_Text:SetFont(f, fs, d, s, o, t, m) self.font = f; self.fsize = fs; self.delta = d; self.shadow = s; self.outline = o; self.thick = t; self.mono = m; self.changedFont = true end
function Widget_Text:SetAlignment(h, v, w) self.justifyH = h; self.justifyV = v; self.wrap = w; self.changedAlignment = true end
function Widget_Text:SetColor(c) self.color = c end
function Widget_Text:SetText(t) self.text = t end

local fontKeys = { font = true, fsize = true, delta = true, shadow = true, outline = true, mono = true }
local alignmentKeys = { justifyH = true, justifyV = true, wrap = true }

-- Generic data access method, checking for changes that require reconfiguration of display objects
function Widget_Text:Set(k, v)
	if self[k] ~= v then -- check if actually changing the data
		self[k] = v
		if fontKeys[k] then self.changedFont = true end
		if alignmentKeys[k] then self.changedAlignment = true end
	end
end

-- Release text when no longer needed, recycle display objects associated with it
function Widget_Text:Release()
	local t = self._display; assert(t, EF("Widget_Text", "undefined text"))
	t.fontstring:Hide(); t.container:Hide()
	table.insert(recycledTexts, t)
	self._display = nil
end

-- Return encoded text flags for outline, thick, mono attributes
function Puffin.TextFlags(outline, thick, mono)
	local t = nil
	if mono then
		if outline then if thick then t = "MONOCHROME,OUTLINE,THICKOUTLINE" else t = "MONOCHROME,OUTLINE" end
		else if thick then t = "MONOCHROME,THICKOUTLINE" else t = "MONOCHROME" end end
	else
		if outline then if thick then t = "OUTLINE,THICKOUTLINE" else t = "OUTLINE" end
		else if thick then t = "THICKOUTLINE" end end
	end
	return t
end

-- Update just the text and color displayed by this widget
function Widget_Text:Update()
	Widget.Update(self)
	local t = self._display; assert(t, EF("Widget_Text", "undefined text"))
	local fs = t.fontstring; assert(fs, EF("Widget_Text", "undefined fontstring"))
	local tc = t.container; assert(tc, EF("Widget_Text", "undefined container"))
	if tc:IsShown() then
		if self.changedAlignment then -- only update alignment properties when changed
			fs:SetJustifyH(self.justifyH or "CENTER"); fs:SetJustifyV(self.justifyV or "MIDDLE"); fs:SetWordWrap(self.wrap)
			self.changedAlignment = false
		end
		if self.changedFont then -- only update font properties when changed
			fs:SetShadowColor(0, 0, 0, self.shadow and 1 or 0); fs:SetShadowOffset(1, -1); fs:SetFontObject(ChatFontNormal)
			local font = self.font
			if font then local mf = media:Fetch("font", font, true); if mf then font = mf end end -- check if text font is in LibSharedMedia
			local fsize = (self.fsize or 10) + (self.delta or 0)
			if font then fs:SetFont(font, fsize, Puffin.TextFlags(self.outline, self.thick, self.mono)) end
			self.changedFont = false
		end
		fs:SetAllPoints()
		local c = self.color; if c then fs:SetTextColor(c.r, c.g, c.b) else fs:SetTextColor(0, 0, 0) end
		if self.text ~= self._text then
			fs:SetText(self.text); self._text = self.text -- only update text when it has changed			
			self.textWidth = fs:GetStringWidth() -- save actual displayed width of the text
		end
		fs:Show()
	else
		fs:Hide(); tc:Hide()
	end
end

-- Widget to show a countdown timer text using a text field (note this is a subclass of Widget_Text)

-- Custom attributes:
-- timeExpire - expiration time
-- timeMax - maximum value of time to display
-- timeLeft - override time computation
-- timeFormat - option format for converting time to text (default: most compact)
-- timeSpaces - option to remove spaces in time formats (default: false)
-- timeCase - option to convert time text characters to lowercase (default: false)

local Widget_Time = Widget_Text:subclass("Widget_Time")
Puffin.Classes.Widget_Time = Widget_Time

local recycledTexts = {}
local textNumber = 0

-- Convert a time value into a compact text string
local TimeFormatOptions = {
	{ 1, 1, 1, 1, 1 }, { 1, 1, 1, 3, 5 }, { 1, 1, 1, 3, 4 }, { 2, 3, 1, 2, 3 }, -- 4
	{ 2, 3, 1, 2, 2 }, { 2, 3, 1, 3, 4 }, { 2, 3, 1, 3, 5 }, { 2, 2, 2, 2, 3 }, -- 8
	{ 2, 2, 2, 2, 2 }, { 2, 2, 2, 2, 4 }, { 2, 2, 2, 3, 4 }, { 2, 2, 2, 3, 5 }, -- 12
	{ 2, 3, 2, 2, 3 }, { 2, 3, 2, 2, 2 }, { 2, 3, 2, 2, 4 }, { 2, 3, 2, 3, 4 }, -- 16
	{ 2, 3, 2, 3, 5 }, { 2, 3, 3, 2, 3 }, { 2, 3, 3, 2, 2 }, { 2, 3, 3, 2, 4 }, -- 20
	{ 2, 3, 3, 3, 4 }, { 2, 3, 3, 3, 5 }, { 3, 3, 3, 2, 3 }, { 3, 3, 3, 3, 5 }, -- 24
	{ 4, 3, 1, 2, 3 }, { 4, 3, 1, 2, 2 }, { 4, 3, 1, 3, 4 }, { 4, 3, 1, 3, 5 }, -- 28
	{ 5, 1, 1, 2, 3 }, { 5, 1, 1, 2, 2 }, { 5, 1, 1, 3, 4 }, { 5, 1, 1, 3, 5 }, -- 32
	{ 3, 3, 3, 2, 2 }, { 3, 3, 3, 3, 4 }, -- 34
}

function Puffin.FormatTime(t, timeFormat, timeSpaces, timeCase)
	if not timeFormat or (timeFormat > #TimeFormatOptions) then timeFormat = 24 end -- default to most compact
	local opt = TimeFormatOptions[timeFormat]
	local h, m, hplus, mplus, s, ts, f
	local o1, o2, o3, o4, o5 = opt[1], opt[2], opt[3], opt[4], opt[5]
	h = math.floor(t / 3600); m = math.floor((t - (h * 3600)) / 60); s = math.floor(t - (h * 3600) - (m * 60))
	hplus = math.floor((t + 3599.99) / 3600); mplus = math.floor((t - (h * 3600) + 59.99) / 60) -- provides compatibility with tooltips
	ts = math.floor(t * 10) / 10 -- truncated to a tenth second
	if t >= 3600 then
		if o1 == 1 then f = string.format("%.0f:%02.0f:%02.0f", h, m, s) elseif o1 == 2 then f = string.format("%.0fh %.0fm", h, m)
			elseif o1 == 3 then f = string.format("%.0fh", hplus) elseif o1 == 4 then f = string.format("%.0fh %.0f", h, m)
			else f = string.format("%.0f:%02.0f", h, m) end
	elseif t >= 120 then
		if o2 == 1 then f = string.format("%.0f:%02.0f", m, s) elseif o2 == 2 then f = string.format("%.0fm %.0fs", m, s)
			else f = string.format("%.0fm", mplus) end
	elseif t >= 60 then
		if o3 == 1 then f = string.format("%.0f:%02.0f", m, s) elseif o3 == 2 then f = string.format("%.0fm %.0fs", m, s)
			else f = string.format("%.0fm", mplus) end
	elseif t >= 10 then
		if o4 == 1 then f = string.format(":%02.0f", s) elseif o4 == 2 then f = string.format("%.0fs", s)
			else f = string.format("%.0f", s) end
	else
		if o5 == 1 then f = string.format(":%02.0f", s) elseif o5 == 2 then f = string.format("%.1fs", ts)
			elseif o5 == 3 then f = string.format("%.0fs", s) elseif o5 == 4 then f = string.format("%.1f", ts)
			else f = string.format("%.0f", s) end
	end
	if not timeSpaces then f = string.gsub(f, " ", "") end
	if timeCase then f = string.upper(f) end
	return f
end

-- Convert time to string for this text widget
local function ConvertTimeToText(w)
	local s = ""
	local timeLeft = w.timeLeft
	if not timeLeft and w.timeExpire then timeLeft = w.timeExpire - GetTime() end
	if timeLeft and timeLeft >= 0 and (not w.timeMax or timeLeft < w.timeMax) then -- check if expiration time reached or timeLeft is longer than maximum limit
		s = Puffin.FormatTime(timeLeft, w.timeFormat, w.timeSpaces, w.timeCase)
	end
	return s, timeLeft
end

-- Create a new text widget object
function Widget_Time:initialize(template, layout)
	self.timeExpire = nil; self.timeMax = nil; self.timeLeft = nil; self.timeFormat = nil; self.timeSpaces = nil; self.timeCase = nil
	Widget_Text.initialize(self, template, layout) -- pass templates and layout to text superclass
	self._callback = Puffin.RegisterCallback("Puffin_Refresh", Widget_Time.Refresh, self, true) -- register suspended event callback
	self._callback_active = false
end

-- Data access methods for setting text-related attributes
function Widget_Time:SetTime(e, m, l, f, s, c)
	self.timeExpire = e; self.timeMax = m; self.timeLeft = l -- these are required fields
	if f then self.timeFormat = f end
	if s then self.timeSpaces = s end
	if c then self.timeCase = c end
end

-- Update just the text and color displayed by this widget
function Widget_Time:Update()
	local timeLeft
	self.text, timeLeft = ConvertTimeToText(self)
	Widget_Text.Update(self)
	local t = self._display
	local tc = t.container
	Puffin.ManageCallback(self, tc:IsShown() and timeLeft and timeLeft >= 0)
end

-- Update the text's animated display objects (i.e., time countdown)
function Widget_Time:Refresh()
	local t = self._display
	local tc = t.container
	local fs = t.fontstring
	if tc:IsShown() and fs then self.text = ConvertTimeToText(self); fs:SetText(self.text) end
end

-- Widget to conditionally show an icon

-- Custom attributes:
-- texture - path to icons's texture (default: color swatch)
-- texcoord - table containing icon's texcoord settings (default: { 0, 1, 0, 1 })
-- texcoordIgnore - don't set icon's texcoords if this is true (default: false)
-- color - color to apply to icon (default: none)

local Widget_Icon = Widget_Backdrop:subclass("Widget_Icon")
Puffin.Classes.Widget_Icon = Widget_Icon

local recycledIcons = {}
local iconNumber = 0 -- used to create unique frame names

-- Create a new text widget object
function Widget_Icon:initialize(template, layout)
	local n, icon = next(recycledIcons) -- get any available recycled widget
	if n then
		recycledIcons[n] = nil
	else
		icon = {}
		iconNumber = iconNumber + 1
		local frameName = "Puffin_Icon" .. tostring(iconNumber) -- generate unique name for icon's frame
		icon.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		icon.texture = icon.container:CreateTexture(nil, "ARTWORK")
	end
	self.texture = nil; self.texcoord = nil; self.color = nil -- clear custom attributes
	Widget_Backdrop.initialize(self, icon, template, layout) -- pass table of display frames used by this widget plus optional template and layout
end

function Widget_Icon:SetTexture(t, tc) self.texture = t; if tc then self.texcoord = tc end end
function Widget_Icon:SetColor(c) self.color = c end
function Widget_Icon:GetIconTexture() return self._display.texture end

-- Release icon when no longer needed, recycle display objects associated with it
function Widget_Icon:Release()
	local t = self._display; assert(t, EF("Widget_Icon", "undefined icon"))
	t.texture:Hide(); t.container:Hide()
	table.insert(recycledIcons, t)
	self._display = nil
end

-- Update just the text and color displayed by this widget
function Widget_Icon:Update()
	Widget_Backdrop.Update(self)
	local t = self._display; assert(t, EF("Widget_Icon", "undefined icon"))
	local tc = t.container; assert(tc, EF("Widget_Icon", "undefined container"))
	local tex = t.texture; assert(tex, EF("Widget_Icon", "undefined texture"))
	if tc:IsShown() then
		local w, h = Pixel(self.width), Pixel(self.height)
		if w > h then w = h elseif h > w then h = w end -- make sure texture is square (with shortest dimension)
		tex:SetSize(w, h)
		if self.texture then tex:SetTexture(self.texture) else tex:SetTexture(1, 1, 1) end
		if not self.texcoordIgnore then
			if self.texcoord then tex:SetTexCoord(unpack(self.texcoord)) else tex:SetTexCoord(0, 1, 0, 1) end
		end
		local c = self.color; if c then tex:SetVertexColor(c.r, c.g, c.b) else tex:SetVertexColor(1, 1, 1) end		
		tex:ClearAllPoints(); tex:SetPoint("CENTER", tc, "CENTER")
		tex:Show()
	else
		tex:Hide(); tc:Hide()
	end
end

-- Widget to display value bars (e.g., health and mana) and timer bars (e.g., auras)

-- Custom attributes:
-- fgTexture - texture to use for the bar foreground (default: color swatch)
-- fgColor -- color to use for the bar foreground (default: blue)
-- fgAlpha - opacity of the bar foreground (default: 1)
-- bgTexture - texture to use for the bar background (default: same as foreground)
-- bgColor - color to use for the bar background (default: same as foreground)
-- bgAlpha - opacity of the bar background (default: 0.6)
-- sparkAlpha - opacity of spark to show at end of bar (default: 0.5)
-- direction - which way foreground bar moves as value decreases, false=right-to-left, true=left-to-right (default: false)
-- fillBars - optionally fill instead of empty the bar (default: false)
-- value - if defined then is the fraction of bar's foreground to fill and timer values are ignored (default: 1)
-- expire - expiration time for timer bar (default: none)
-- duration - total time that the timer bar will run (default: none)
-- timeLeft - if set then override calculation from expiration time
-- maxTime - maximum time actually displayed on timer bar (default: none)

local Widget_Bar = Widget_Backdrop:subclass("Widget_Bar")
Puffin.Classes.Widget_Bar = Widget_Bar

local recycledBars = {}
local barNumber = 0

-- Create a new bar widget object
function Widget_Bar:initialize(template, layout)
	local n, bar = next(recycledBars) -- get any available recycled widget
	if n then
		recycledBars[n] = nil
	else
		bar = {}
		barNumber = barNumber + 1
		local frameName = "Puffin_Bar" .. tostring(barNumber) -- generate unique name for bar's frame
		bar.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		bar.bgTexture = bar.container:CreateTexture(nil, "BACKGROUND")
		bar.fgTexture = bar.container:CreateTexture(nil, "BORDER")	
		bar.spark = bar.container:CreateTexture(nil, "OVERLAY")
		bar.spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
		bar.spark:SetBlendMode("ADD"); bar.spark:SetTexCoord(0, 1, 0, 1)
	end
	self.changedBars = true
	self.fgTexture = nil; self.fgColor = nil; self.fgAlpha = 1; self.bgTexture = nil; self.bgColor = nil; self.bgAlpha = 0.6
	self.direction = false; self.sparkAlpha = 0.5; self.fillBars = false
	self.value = nil; self.expire = nil; self.duration = nil; self.timeLeft = nil; self.maxTime = nil
	Widget_Backdrop.initialize(self, bar, template, layout) -- pass table of display frames used by this widget plus optional template and layout
	self._callback = Puffin.RegisterCallback("Puffin_Refresh", Widget_Bar.Refresh, self, true) -- register suspended event callback	
end

-- Configure with any changes to template and layout
function Widget_Bar:Configure(template, layout)
	self.changedBars = true
	Widget_Backdrop.Configure(self, template, layout)
end

-- Data access method for setting time-related attributes
function Widget_Bar:SetTime(e, d, l, m) self.expire = e; self.duration = d; self.timeLeft = l; self.maxTime = m end

-- Data access method for setting bar textures and related attributes
function Widget_Bar:SetBars(ft, fc, fa, bt, bc, ba)
	self.fgTexture = ft; self.fgColor = fc; self.fgAlpha = fa; self.bgTexture = bt; self.bgColor = bc; self.bgAlpha = ba; self.changedBars = true
end

-- Data access method for setting value for non-timer bar
function Widget_Bar:SetValue(v) self.value = v end

-- Data access method for bar configuration options
function Widget_Bar:SetOptions(d, sa, fb, fp) self.direction = d; self.sparkAlpha = sa; self.fillBars = fb; self.changedBars = true end

local barKeys = { fgTexture = true, fgColor = true, fgAlpha = true, bgTexture = true, bgColor = true, bgAlpha = true, sparkAlpha = true, width = true, height = true }

-- Generic data access method, checking for changes that require reconfiguration of display objects
function Widget_Bar:Set(k, v)
	if self[k] ~= v then -- check if actually changing the data
		self[k] = v
		if barKeys[k] then self.changedBars = true end
	end
end

-- Release a bar when no longer needed, recycle display objects associated with it
function Widget_Bar:Release()
	local t = self._display; assert(t, EF("Widget_Bar", "undefined bar"))
	t.fgTexture:Hide(); t.bgTexture:Hide(); t.spark:Hide(); t.container:Hide()
	table.insert(recycledBars, t)
	self._display = nil
end

-- Compute fill fraction for a value or timer bar and adjust foreground texture and spark accordingly
function Widget_Bar:FillBar(tc, tfg, tsp)
	local remaining
	local fill = 1
	if self.value then -- value bars specify a fraction to show
		fill = self.value
	elseif self.duration and self.expire then -- timer bars require calculation to figure out fraction to show
		local maxTime = self.maxTime or self.duration
		remaining = self.timeLeft or (self.expire - GetTime()) -- calculate remaining time in seconds, unless timeLeft set as an override
		if remaining < 0 then remaining = 0 end -- make sure no rounding funnies
		if remaining > self.duration then remaining = self.duration end -- and no inaccurate durations!
		if remaining < maxTime then fill = remaining / maxTime end -- calculate fraction of time remaining
		-- Puffin.Debug("fill", remaining, self.expire, self.duration, self.timeLeft)
		if self.fillBars then fill = 1 - fill end
	end
	local fw = Pixel(self.width * fill) -- doesn't matter if absolutely exact
	if fw >= 0.25 then -- but must be aligned prior to this test!
		local fgAnchor, sparkAnchor
		tfg:SetSize(fw, tc:GetHeight())
		if self.direction then
			fgAnchor = "RIGHT"; sparkAnchor = "LEFT"; tfg:SetTexCoord(fill, 0, fill, 1, 0, 0, 0, 1)
		else
			fgAnchor = "LEFT"; sparkAnchor = "RIGHT"; tfg:SetTexCoord(0, 0, 0, 1, fill, 0, fill, 1)
		end
		tfg:ClearAllPoints(); tfg:SetPoint(fgAnchor, tc, fgAnchor); tfg:Show()
		tsp:ClearAllPoints(); tsp:SetPoint("CENTER", tfg, sparkAnchor, 0, Pixel(-1))
		if self.sparkAlpha and self.sparkAlpha > 0 and fill < 1 then tsp:Show() else tsp:Hide() end
	else tfg:Hide(); tsp:Hide() end
	return remaining
end

-- Update the bar's value and color
function Widget_Bar:Update()
	Widget_Backdrop.Update(self)
	local t = self._display; assert(t, EF("Widget_Bar", "undefined bar"))
	local tc = t.container; assert(tc, EF("Widget_Bar", "undefined container"))
	local tfg, tbg, tsp = t.fgTexture, t.bgTexture, t.spark
	local timeLeft
	
	if tc:IsShown() then
		local w, h = Pixel(self.width), Pixel(self.height)	
		if self.changedBars then
			local tex = self.bgTexture or self.fgTexture
			local c = self.bgColor or self.fgColor
			if tex then mt = media:Fetch("statusbar", tex, true); if mt then tex = mt end end -- check if background texture is in LibSharedMedia
			if tex then tbg:SetTexture(tex) else tbg:SetTexture(1, 1, 1) end
			if not c then c = defaultBlue end; tbg:SetVertexColor(c.r, c.g, c.b, 1)
			tbg:SetSize(w, h); tbg:SetTexCoord(0, 1, 0, 1); tbg:SetAlpha(self.bgAlpha or 0.6)
			tbg:ClearAllPoints(); tbg:SetPoint("CENTER", tc, "CENTER")

			if self.bgTexture ~= self.fgTexture then -- if same then already have texture for foreground
				tex = self.fgTexture -- if not same then look up foreground texture
				if tex then mt = media:Fetch("statusbar", tex, true); if mt then tex = mt end end -- check if foreground texture is in LibSharedMedia
			end
			if tex then tfg:SetTexture(tex) else tfg:SetTexture(1, 1, 1) end
			if self.fgColor then local c = self.fgColor; tfg:SetVertexColor(c.r, c.g, c.b, 1) end
			tfg:SetAlpha(self.fgAlpha or 1)
			tsp:SetSize(Pixel(16), h * 1.5); tsp:SetAlpha(self.sparkAlpha or 0)
			self.changedBars = false
		end
			
		if self.bgAlpha ~= 0 then tbg:Show() else tbg:Hide() end
		if self.fgAlpha ~= 0 then timeLeft = self:FillBar(tc, tfg, tsp) end
	else
		tfg:Hide(); tbg:Hide(); tsp:Hide(); tc:Hide()
	end
	Puffin.ManageCallback(self, tc:IsShown() and not self.value)
end

-- Update the bar's animated display objects (i.e., flashing, foreground bar and spark, overlay and artwork bars)
function Widget_Bar:Refresh()
	if not self.hidden and (self.condition ~= false) then
		local t = self._display
		if self.fgAlpha ~= 0 then self:FillBar(t.container, t.fgTexture, t.spark) end
	end
end

-- Widget to display rectangles with optional borders and flashing

-- Custom attributes:
-- texture - path to rectangle's texture (default: color swatch)
-- texcoord - table containing texture's texcoord settings (default: { 0, 1, 0, 1 })
-- color - color to apply to texture (default: none)
-- flashPeriod - if defined and greater than 0 then flash rectangle over that number of seconds
-- valueWidth - fraction of rectangle's width to fill (default: 1)
-- valueHeight - fraction of rectangle's height to fill (default: 1)
-- valueOffsetX - fraction of rectangle's width to offset horizontal position (default: 0)
-- valueOffsetY - fraction of rectangle's height to offset vertical position (default: 0)
-- valueAlpha - alpha setting for the rectangles interior (default: 1)
-- minimumWidth - minimum width in pixels to fill (default: 0.5)
-- minimumHeight - minimum height in pixels to fill (default: 0.5)

local Widget_Rectangle = Widget_Backdrop:subclass("Widget_Rectangle")
Puffin.Classes.Widget_Rectangle = Widget_Rectangle

local recycledRectangles = {}
local rectangleNumber = 0

-- Create a new rectangle widget object
function Widget_Rectangle:initialize(template, layout)
	local n, rect = next(recycledRectangles) -- get any available recycled widget
	if n then
		recycledRectangles[n] = nil
	else
		rect = {}
		rectangleNumber = rectangleNumber + 1
		local frameName = "Puffin_Rectangle" .. tostring(rectangleNumber) -- generate unique name for rect's frame
		rect.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		rect.texture = rect.container:CreateTexture(nil, "BACKGROUND")
	end
	self.texture = nil; self.texcoord = nil; self.color = nil; self.flashPeriod = nil
	self.valueWidth = nil; self.valueHeight = nil; self.valueOffsetX = nil; self.valueOffsetY = nil; self.valueAlpha = nil
	self.minimumWidth = nil; self.minimumHeight = nil
	Widget_Backdrop.initialize(self, rect, template, layout) -- pass table of display frames used by this widget plus optional template and layout
	self._callback = Puffin.RegisterCallback("Puffin_Refresh", Widget_Rectangle.Refresh, self, true) -- register suspended callback
end

-- Data access method for setting texture and color-related attributes
function Widget_Rectangle:SetRectangle(t, tc, c, fp) self.texture = t; self.texcoord = tc; self.color = c; self.flashPeriod = fp end

-- Data access method for setting value-related attributes
function Widget_Rectangle:SetValue(vw, vh, vx, vy, va, mw, mh)
	self.valueWidth = vw; self.valueHeight = vh; self.valueOffsetX = vx; self.valueOffsetY = vy; self.valueAlpha = va
	self.minimumWidth = mw; self.minimumHeight = mh
end

-- Release a bar when no longer needed, recycle display objects associated with it
function Widget_Rectangle:Release()
	local t = self._display; assert(t, EF("Widget_Rectangle", "undefined rect"))
	t.texture:Hide(); t.container:Hide()
	table.insert(recycledRectangles, t)
	self._display = nil
end

-- Calculate alpha for flashing, period is how long the total flash time should last
local function FlashAlpha(maxAlpha, period)
	if not period or period == 0 then return 1 end
	local frac = GetTime() / period
	frac = frac - math.floor(frac) -- get fractional part of current period
	if frac >= 0.5 then frac = 1 - frac end -- now goes from 0 to 0.5 then back to 0
	return (maxAlpha / 2) + (maxAlpha * frac) -- drops to half of maxAlpha
end

-- Update the widget's display objects based on current data attribute values (or defaults)
function Widget_Rectangle:Update()
	Widget_Backdrop.Update(self)
	local t = self._display; assert(t, EF("Widget_Rectangle", "undefined rect"))
	local tc = t.container; assert(tc, EF("Widget_Rectangle", "undefined container"))
	local tex = t.texture; assert(tex, EF("Widget_Rectangle", "undefined texture"))
	local vw, vh = Pixel(math.floor((self.valueWidth or 1) * self.width)), Pixel(math.floor((self.valueHeight or 1) * self.height))
	
	if tc:IsShown() and vw >= (self.minimumWidth or 0.5) and vh >= (self.minimumHeight or 0.5) then
		if self.texture then tex:SetTexture(self.texture) else tex:SetTexture(1, 1, 1) end -- plain color block
		if self.texcoord then tex:SetTexCoord(unpack(self.texcoord)) else tex:SetTexCoord(0, 1, 0, 1) end
		local c = self.color
		if not c then c = defaultBlack end; tex:SetVertexColor(c.r, c.g, c.b, 1)			
		local alpha = self.alpha or 1 -- flashing only adjusts foreground alpha
		if self.flashPeriod then alpha = FlashAlpha(alpha, self.flashPeriod) end -- adjust alpha if flashing
		
		local vx, vy = (self.valueOffsetX or 0) * self.width, (self.valueOffsetY or 0) * self.height
		tex:ClearAllPoints(); tex:SetPoint("BOTTOMLEFT", tc, "BOTTOMLEFT", Pixel(vx), Pixel(vy))
		tex:SetAlpha(self.valueAlpha or 1); tex:SetSize(vw, vh); tex:Show()
	else
		tex:Hide(); tc:Hide()
	end
	Puffin.ManageCallback(self, tc:IsShown() and self.flashPeriod)
end

-- Run flash animation
function Widget_Rectangle:Refresh()
	if not self.hidden and (self.condition ~= false) then
		local t = self._display
		local vw, vh = self.valueWidth or 1, self.valueHeight or 1
		if self.flashPeriod and vw > 0 and vh > 0 then
			local alpha = FlashAlpha(self.alpha or 1, self.flashPeriod) -- adjust alpha if flashing
			t.container:SetAlpha(alpha)
		end
	end
end
